(*
 *	Copyright (c) 2002 by Laboratoire Spcification et Vrification (LSV),
 *	CNRS UMR 8643 & ENS Cachan.
 *	Written by Jean Goubault-Larrecq.  Not derived from licensed software.
 *
 *	Permission is granted to anyone to use this software for any
 *	purpose on any computer system, and to redistribute it freely,
 *	subject to the following restrictions:
 *
 *	1. Neither the author nor its employer is responsible for the 
 *         consequences of use of this software, no matter how awful, 
 *         even if they arise from defects in it.
 *
 *	2. The origin of this software must not be misrepresented, either
 *	   by explicit claim or by omission.
 *
 *	3. Altered versions must be plainly marked as such, and must not
 *	   be misrepresented as being the original software.
 *
 *	4. This software is restricted to non-commercial use only.  Commercial
 *	   use is subject to a specific license, obtainable from LSV.
 *	   This was developed as part of the EVA project; as such, access
 *	   to this software by partners of the RNTL EVA Project (Trusted
 *	   Logic S.A., Verimag) is defined by the intellectual property
 *	   agreement signed as part of this project.
 *)

(*
 * Pretty print specification in abstract syntax into CPL
 *)

open Format;;

open Trace;;
open Abstract;;
open Termutil;;
open Types;;
open Pp;;


(***************************************************************
 * Separators for pp_list
 ***************************************************************)

(* standard comma separation to use for pp_list *)
let comma_sep () = 
  print_string ","; print_space();;

(* standard blank space separation to use for pp_list *)
let space_sep () = 
  print_space();;

(* standard semicolumn separation to use for pp_list *)
let semicol_sep () = 
  print_string ";"; print_space();;

(* standard null separation to use for pp_list *)
let null_sep () = ();;

(* standard new line separation to use for pp_list *)
let newline_sep () = 
  print_newline ();;

(***************************************************************
 * lisp like printers
 ***************************************************************)
let open_list name =
  open_box 0;
  print_string "(";
  print_string name;;
  (* print_newline (); *)
  (* open_box 2;; *)
  
let close_list () =
  close_box ();
  print_string ")";;

let print_id name = 
  print_string "\"";
  print_string name;
  print_string "\"";;

(***************************************************************
 * specific to CPL
 ***************************************************************)

let rec has_typedcl = function
    [] -> false
  | d::l -> (
      match d with
        EVA_DCL_ID (_,_,_) 
      | EVA_DCL_FUN (_,_,_,_,_,_)
      | EVA_DCL_KEYPAIR (_,_,_,_,_,_) -> true
      | _ -> has_typedcl l);;

let rec string_of_label = function
    EVA_NOLABEL -> "NOLABEL"
  | EVA_LABEL s -> s
and string_of_llabel { lla_label = l } =
  string_of_label l;;


(***************************************************************
 * Pretty printing functions
 ***************************************************************)

(* for identifiers *)
let cplpp_string s = 
  print_string "\"";
  print_string s;
  print_string "\"";;

let rec cplpp_label = function
    EVA_NOLABEL -> ()
  | EVA_LABEL s -> 
	cplpp_string s
and cplpp_llabel { lla_label = l } =
  cplpp_label l;;

let rec cplpp_type = function
    EVA_VOID -> cplpp_string "VOID"
  | EVA_PRINCIPAL -> cplpp_string "principal"
  | EVA_NUMBER -> cplpp_string "number"
  | EVA_AALGO -> cplpp_string "asym_algo"
  | EVA_SALGO -> cplpp_string "sym_algo"
  | EVA_TALGO -> cplpp_string "*algo*"
  | EVA_USERTYPE s -> cplpp_string s
  | EVA_BASETYPE s -> cplpp_string s
  | EVA_FUN t -> cplpp_string "fun_"; cplpp_type t
  | EVA_HFUN t -> cplpp_string "hfun_"; cplpp_type t
  | EVA_KEY t -> cplpp_string "key_"; cplpp_type t
and cplpp_ltype { lty_type = t } =
  cplpp_type t;;

let rec cplpp_scope = function
    EVA_CST -> cplpp_string "constant"
  | EVA_PARAM -> cplpp_string "parameter"
  | EVA_PRIVATE -> cplpp_string "private";
and cplpp_lscope { lsc_scope = s } =
  cplpp_scope s;;

let rec cplpp_id (EVA_ID s) =
  cplpp_string s
and cplpp_lid { lid_id = i }  =
  cplpp_id i;;

let rec cplpp_value = function
    EVA_INTRUDER -> cplpp_string "I"
  | EVA_VALUE s -> cplpp_string s
and cplpp_lvalue { lva_value = v } =
  cplpp_value v;;

let rec cplpp_term term =
  begin
    open_box 0;
    (match term with
         (* EVA_TERM_VOID -> print_string "" *)
         EVA_TERM_ID s -> 
	   cplpp_string s
       | EVA_TERM_CST s -> 
	   cplpp_string s
       | (EVA_TERM_VALUE v) -> 
	   cplpp_value v
       | (EVA_TERM_ALIAS (a, ta)) -> 
	   cplpp_string a
       | (EVA_LOCATED (i, r)) -> 
	   open_list "located";
	   cplpp_string "_"; (* label *)
           cplpp_lid i; 
           cplpp_lid r; 
	   close_list ();
       | (EVA_QUANTIFIED (i, _)) -> 
	   print_string i
       | EVA_P t -> 
           open_list "p";
           cplpp_lterm t; 
           close_list ()
       | EVA_A t -> 
           open_list "a";
           cplpp_lterm t; 
           close_list ()
       | EVA_SA t -> 
           open_list "sa";
           cplpp_lterm t; 
           close_list ()
       | EVA_AA t -> 
           open_list "aa";
           cplpp_lterm t; 
           close_list ()
       | EVA_U (i, t) -> 
           open_list ("u_"^i);
           cplpp_lterm t; 
           close_list ()
       | EVA_CONS tl -> 
	   open_list "tuple";
	   pp_list cplpp_lterm space_sep tl;
	   close_list ();
       | EVA_APP (i, etl) -> 
	   cplpp_app i etl
       | EVA_CRYPT (a, k, t) 
       | EVA_SIGN (a, k, t) -> 
	   cplpp_cipher a k t 
       | EVA_PCENT (t1, t2) -> 
	   raise (Internal_error "cplpp_term (%)")
    );
    close_box ();
  end
and cplpp_lterm { lte_term = t } =
  cplpp_term t
(* app_root = i; app_args = etl *)
and cplpp_app i etl =
    begin
      open_list "apply";
      cplpp_lid i;
      print_space ();
      pp_list cplpp_lterm space_sep etl;
      close_list ();
    end
(* crypt_algo = a; crypt_key = k; crypt_content = t *)
and cplpp_cipher a k t =
   begin 
     open_list "crypt";
     cplpp_algo a;
     print_space ();
     cplpp_lterm t; 
     print_space ();
     cplpp_lterm k;
     close_list ();
   end
and cplpp_algo = function
    EVA_VANILLA -> 
      open_list "vanilla";
      close_list ();
  | (EVA_ALGO t) -> 
      cplpp_lterm t;;

(* returns a new artificial id for representing both keys of the pair *)
let keypair_id k1 k2 =
  let kpid = ("_"^(string_of_lid k1)^(string_of_lid k2))
  in { k1 with lid_id = (EVA_ID kpid) };;

(* know_principal = i; know_contents = tl *)
let rec cplpp_dcl_know = function
    EVA_DCL_KNOW (i, tl) ->
      open_list "knows"; 
      pp_list cplpp_lterm comma_sep tl; 
      close_list ()
  | _ -> ()
and cplpp_ldcl_know { ldc_dcl = d } =
  cplpp_dcl_know d;;

(* return the declaration knows for the given role in 
   the given declaration list *)
let rec find_dcl_knows i = function
    [] -> raise (Internal_error "find_dcl_knows : not found")
  | d::l -> (
      match d with
	{ ldc_dcl = (EVA_DCL_KNOW (r, _)) } -> 
	  if (lid_equal r i)
	  then d
	  else find_dcl_knows i l
      | _ -> find_dcl_knows i l);;

(* print id type declarations, ignore others *)
let rec cplpp_dcl_type = function
    EVA_DCL_ID (ibl, s, t) -> 
      (List.iter (fun x -> cplpp_dcl_id x s t; print_newline ()) ibl)
  | EVA_DCL_FUN (i, s, t, tl, bh, bs) ->  
      cplpp_dcl_fun i s t tl bh bs
  | EVA_DCL_KEYPAIR (a, k1, k2, s, t, tl) -> 
      cplpp_dcl_keypair a k1 k2 s t tl
  | _ -> ()
and cplpp_ldcl_type { ldc_dcl = d } =
  cplpp_dcl_type d
and cplpp_dcl_id (i, f) s t =
  begin
    open_list "type";
    cplpp_lid i; 
    print_space ();
    cplpp_lscope s;
    print_space ();
    cplpp_ltype t;
    if f
    then (print_space (); cplpp_string "fresh";);
    close_list ();
  end
and cplpp_dcl_fun i s t tl bh bs =
  begin
    open_list "type";
    cplpp_lid i; 
    print_space ();
    cplpp_lscope s;
    print_space ();
    (let funkind =
      if ( bh && bs )
      then "secret-one-way-function"
      else if bh
      then "one-way-function"
      else if bs
      then "secret-function"
      else "function" in 
     open_list funkind;
    cplpp_ltype t;
    print_space ();
    pp_list cplpp_ltype space_sep tl;
    close_list ());
    close_list ();
  end
(* declaration of a one-way function representing both keys of the keypair.
 * The 2 keys are defined after as aliases *
 * At this point, the algo is ignored *)
and cplpp_dcl_keypair a k1 k2 s t tl =
  begin
    open_list "type";
    cplpp_lid (keypair_id k1 k2);
    print_space ();
    cplpp_lscope s;
    print_space ();
    (* the keys are always assumed hash *)
    open_list "one-way-function";
    cplpp_ltype t;
    print_space ();
    pp_list cplpp_ltype space_sep tl;
    close_list ();
    close_list ()
  end;;

let rec cplpp_dcl_alias = function
    EVA_DCL_ALIAS (i, t) -> 
      begin 
        open_list "alias"; 
        cplpp_lid i;
	print_space (); 
        cplpp_lterm t; 
	close_list ();
      end
  | _ -> () 
and cplpp_ldcl_alias { ldc_dcl = d } =
  cplpp_dcl_alias d;;


(* print two aliases decl. corr. to the given keypair declaration 
 * do nothing if the declaration is not keypair decl. *)
let rec cplpp_dcl_keypair_alias = function
    EVA_DCL_KEYPAIR (a, k1, k2, s, t, tl) -> 
      begin 
	(* first alias for k1 *)
        open_list "alias"; 
        cplpp_lid k1;
	print_space (); 
        open_list "lambda_pubk"; 
	cplpp_algo a; 
	print_space (); 
	cplpp_lid (keypair_id k1 k2); 
	close_list ();
	close_list ();
	print_newline ();
	(* second alias for k2 *)
        open_list "alias"; 
        cplpp_lid k2;
	print_space (); 
        open_list "lambda_privk"; 
	cplpp_algo a; 
	print_space (); 
	cplpp_lid (keypair_id k1 k2); 
	close_list ();
	close_list ();
      end
  | _ -> ()
and cplpp_ldcl_keypair_alias { ldc_dcl = d } =
  cplpp_dcl_keypair_alias d;;
    
let rec cplpp_dcl_axiom = function
    EVA_DCL_AXIOM (t1, t2, t, ql) -> 
      begin 
        open_list "axiom";
        cplpp_lterm t1;
        print_space ();
        cplpp_lterm t2;
        print_space ();
        cplpp_type t;
        print_space (); 
        pp_list cplpp_optional_quantification space_sep ql;
        close_list ();
      end
  | _ -> ()
and cplpp_ldcl_axiom { ldc_dcl = d } =
  cplpp_dcl_axiom d
and cplpp_optional_quantification (i, t) =
  begin 
    open_list "type";
    cplpp_lid i; 
    print_space ();
    cplpp_string "quantified";
    print_space ();
    cplpp_ltype t;
    close_list ();
  end;;


let rec cplpp_dcl_value = function
    EVA_DCL_VALUE (vl, t) ->
      (List.iter (fun x -> cplpp_dcl_singlevalue x t; print_newline ()) vl)
  | _ -> ()
and cplpp_ldcl_value { ldc_dcl = d } =
  cplpp_dcl_value d
and cplpp_dcl_singlevalue v t = 
  begin 
    open_list "value-type";
    (* should not be INTRUDER *)
    cplpp_lvalue v;
    print_space ();
    cplpp_ltype t;
    close_list ();
  end;;



(* instructions, processes *)

(* prints the given intruction i for the program of the 
   given role (located id) r *)
let rec cplpp_instr r i = 
  (
   match i with
     EVA_SKIP -> print_space ()
   | EVA_MSG (l, is, ir, t) -> 
       if (lid_equal r is)
       then 
	 begin
	   open_list "send";
	   cplpp_lterm t;
	   close_list ();
	 end
       else if (lid_equal r ir)
       then
	 begin
	   open_list "recv";
	   cplpp_lterm t;
	   close_list ();
	 end
       else
	 raise (Internal_error "cplpp_instr: msg")
   | EVA_ASSIGN (ir, iv, t) -> 
       if (lid_equal ir r)
       then
	 begin
	   open_list "assign";
	   cplpp_lid iv;
	   cplpp_lterm t
	 end
       else 
	 raise (Internal_error "cplpp_instr: assign")
   | EVA_COMP (ir, iv, t) ->
       if (lid_equal ir r)
       then
	 begin
	   open_list "comp";
	   cplpp_lid iv;
	  cplpp_lterm t
	 end
       else 
	 raise (Internal_error "cplpp_instr: compare")
   | EVA_BLOCK b -> 
       raise (Internal_error "cplpp_instr: block (not supported in CPL")
      (* pp_list cplpp_linstr null_sep b;  *)
   | EVA_SWITCH (t, cl) -> 
       raise (Internal_error "cplpp_instr: switch (not supported in CPL")
  );
  print_newline ();
and cplpp_linstr r { lin_instr = i } =
  cplpp_instr r i;;

(* param: 
   - located decl list (to extract knows decl) 
   - protocol in the form EVA_PROCESS program list  *)
let rec cplpp_protocol dl = function
    (EVA_MP l) -> 
      raise (Internal_error "cplpp_protocol: MP (not supported in CPL)")
  | (EVA_PROCESS pl) -> 
      open_list "system";
      print_newline (); 
      open_box 2;
      pp_list (cplpp_process dl) newline_sep pl;
      print_newline ();
      close_box ()     
and cplpp_lprotocol dl { lpr_protocol = p } =
  cplpp_protocol dl p
and cplpp_process dl { pr_role = r; pr_privates = il; pr_messages = b } =
  begin 
    open_box 0;
    open_list "repeat-process"; 
    print_newline ();
    open_box 2;
    open_list "private";
    pp_list cplpp_lid space_sep il;
    close_list ();
    print_newline ();
    cplpp_lid r;
    print_newline ();
    cplpp_string (get_first_label b);
    print_newline ();
    cplpp_ldcl_know (find_dcl_knows r dl);
    print_newline ();
    cplpp_transitionlist r (get_first_label b) b;
  end
and get_first_label = function
    [] -> "begin";
  | i::l -> (
      match i.lin_instr with
	EVA_MSG (l, _, _, _) -> (string_of_llabel l)
      | _ -> "begin")
and cplpp_transitionlist r label = function
    [] -> ()
  | { lin_instr = i1 }::({ lin_loc = loc2; lin_instr = i2 }::l) -> 
      let (l1, l2) = (labels_of_instrs label (i1, i2)) in
      begin
	open_list "trans";
	cplpp_string l1;
	print_space ();
	cplpp_string l2;
	print_space ();
	cplpp_instr r i1;
	close_list ();
	print_newline ();
	cplpp_transitionlist r l2 ({ lin_loc = loc2; lin_instr = i2 }::l)
      end
    (* singleton list (last linstr) *)
  | { lin_instr = i }::[] -> 
      open_list "trans";
      cplpp_string label;
      print_space ();
      cplpp_string "end.";
      print_space ();
      cplpp_instr r i;
      close_list ();
(* only MSG have label, create label for the others *)
and labels_of_instrs label = function
    ((EVA_MSG (l1, _, _, _)), (EVA_MSG (l2, _, _, _))) ->
      ((string_of_llabel l1), (string_of_llabel l2))
  | ((EVA_MSG (l1, _, _, _)), _) ->
      ((string_of_llabel l1), ((string_of_llabel l1)^".1_"))
  | ( _, (EVA_MSG (l2, _, _, _))) ->
      (label, (string_of_llabel l2))
  | (_, _) ->
      (label, (label^".2_"));;
	  

(* sessions *)

let rec cplpp_constraint = function
    EVA_EQ_CONSTRAINT (i1, i2) -> 
      open_list "cons_eq";
      cplpp_lid i1; 
      print_space ();
      cplpp_lid i2; 
      close_list ()
  | EVA_NEQ_CONSTRAINT (i1, i2) -> 
      open_list "cons_neq";
      cplpp_lid i1; 
      print_space ();
      cplpp_lid i2; 
      close_list ()
  | EVA_MEMBERSHIP_CONSTRAINT (i, tl) -> 
      open_list "cons_member";
      cplpp_lid i; 
      print_space ();
      pp_list cplpp_lterm space_sep tl; 
      close_list ()
  | EVA_DOMAIN_CONSTRAINT (i, s) -> 
      open_list "cons_domain";
      cplpp_lid i; 
      print_space ();
      print_string s;
      close_list ()
and cplpp_lconstraint { lco_constraint = c } =
  cplpp_constraint c;;

let rec cplpp_session = function
    EVA_BANG -> 
      open_list "bang";
      close_list ()
  | EVA_CONSTRAINED_SESSIONS l -> 
      open_list "constrained_session"; 
      pp_list cplpp_lconstraint space_sep l; 
      close_list (); 
  | EVA_SESSIONS al -> 
      open_list "session_bang";
      pp_list cplpp_sassignment space_sep al; 
      close_list ();
  | EVA_SESSION (l, al) -> 
      open_list "session";
      cplpp_llabel l; print_space (); 
      pp_list cplpp_sassignment space_sep al; 
      close_list ();
and cplpp_lsession { lse_session = s } = 
  cplpp_session s
and cplpp_sassignment (i, v) = 
  open_list "env";
  cplpp_lid i; print_space(); 
  cplpp_lterm v; 
  close_list ();;


(* assume, claim *)

let rec cplpp_atom = function
    EVA_TRUE -> 
      open_list "true"; 
      close_list ()
  | EVA_FALSE -> 
      open_list "false"; 
      close_list ()
  | EVA_ATOM_EQ (t1, t2) -> 
      open_list "eq"; 
      cplpp_lterm t1; print_space ();
      cplpp_lterm t2; 
      close_list ()
  | EVA_ATOM_NEQ (t1, t2) -> 
      open_list "neq"; 
      cplpp_lterm t1; print_space ();
      cplpp_lterm t2; 
      close_list ()
  | EVA_ATOM_HONEST t -> 
      open_list "honest"; 
      cplpp_lterm t;
      close_list ()
  | EVA_ATOM_SECRET  t -> 
      open_list "secret"; 
      cplpp_lterm t;
      close_list ()
  | EVA_ATOM_USER (i, t) -> 
      open_list (string_of_lid i);
      cplpp_lterm t;
      close_list ()     
and cplpp_latom { lat_atom = a } =
  cplpp_atom a;;


let rec cplpp_assume = function
    EVA_ASSUME (ql, al, a) ->
      if (a.lat_atom = EVA_FALSE)
      then open_list "quantified_negative"
      else open_list "quantified_clause";
      open_list "forall";
      pp_list cplpp_optional_quantification space_sep ql;	   
      close_list ();
      print_space ();
      pp_list cplpp_latom space_sep al;
      if (a.lat_atom <> EVA_FALSE)
      then print_space (); cplpp_latom a;
      close_list ();
      print_newline ();      
  | _ -> ()
and cplpp_claim = function
    EVA_CLAIM (l, al, a) ->
      if (a.lat_atom = EVA_FALSE)
      then open_list "negative"
      else open_list "clause";
      pp_list cplpp_latom space_sep al;
      if (a.lat_atom <> EVA_FALSE)
      then print_space (); cplpp_latom a;
      close_list ();
      print_newline ();
  | _ -> ()
and cplpp_lassume { lst_statement = s } =
  cplpp_assume s
and cplpp_lclaim { lst_statement = s } =
  cplpp_claim s;;

let cplpp_spec = function 
    { sp_label = l;
      sp_declarations = dl;
      sp_protocol = p;
      sp_valuedcl = vdl;
      sp_sessions = sl;
      sp_statements = stl } ->
        open_list "compiled-spec";
        print_newline ();
        open_box 2;
        open_list "types";
        print_newline ();
        open_box 2; 
        pp_list cplpp_ldcl_type newline_sep dl;
        pp_list cplpp_ldcl_value newline_sep dl;	
        close_box ();
        close_list (); (* end types *)
        print_newline ();        
        open_list "values";
        print_newline ();
        open_box 2;
        pp_list cplpp_ldcl_alias newline_sep dl;
        print_newline ();
        (* new aliases for keypairs *) 
        pp_list cplpp_ldcl_keypair_alias newline_sep dl;
        close_box ();
        close_list (); (* end values *)
        print_newline ();
	open_list "axioms";
	print_newline ();
        open_box 2;
        pp_list cplpp_ldcl_axiom newline_sep dl;
        close_box ();
        close_list (); (* end values *)
        print_newline ();
	(* system *)
        cplpp_lprotocol dl p;
        print_newline ();
	open_list "sessions";
	print_newline ();
        open_box 2;
        pp_list cplpp_lsession newline_sep sl;
	close_box ();
	close_list ();
        print_newline ();
	open_list "assume";
	open_box 2;
        pp_list cplpp_lassume newline_sep stl;
	close_box ();
	close_list ();
        print_newline ();
	open_list "claim";
	open_box 2;
        pp_list cplpp_lclaim newline_sep stl;
	close_box ();
	close_list ();
        print_newline ();;

